/**
 * 2x2 Matrix Visualization
 *
 * This is a program to visualize a 2x2 matrix and its equivalent two 
 * dimensional linear transformation. 
 *
 * It displays eigenvalues, eigenvectors, and the determinant.
 *
 * Made with Processing (https://processing.org/)
 *
 * @author Sarah Greer
 * http://sygreer.com
 *
 *
 * This code is licensed under the GNU General Public License version 3 (GPLv3)
 * More information can be found here: 
 * https://www.gnu.org/licenses/gpl-3.0.en.html
 * 
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or (at your option) any later 
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with 
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

// initialize plots and buttons
Plot plot;              // plot for matrix
Button r;               // reset button
Button submit;          // submit button

Matrix mat;             // matrix being displayed
TextBox A, B, C, D;     // text boxes of matrix

// selected
int selected_prev = 0;
int selected;

// zoom
int zoom_level = 0;
Button zin, zout;

private float xmin, xmax, ymin, ymax;   // current zoom value
private float xmin_orig, xmax_orig, ymin_orig, ymax_orig; // default zoom values

void setup(){
  size(800,800);
  background(255);

  // default zoom
  xmin = ymin = -2;
  xmax = ymax = 2;

  xmin_orig = xmin;
  xmax_orig = xmax;
  ymin_orig = ymin;
  ymax_orig = ymax;

  // initiaize plot, matrix, and buttons
  plot = new Plot(800,800,xmin, ymin, xmax, ymax, #d6ebf2);
  mat = new Matrix(plot, 1, 0, 0, 1);

  zin = new Button(50, 50, 50, 50, #999999);
  zout = new Button(50, 100, 50, 50, #999999);

  zin.setText(20, "+", #ffffff);
  zout.setText(20, "-", #ffffff);

  frameRate(10);
}

void draw(){
  background(#ffffff);
  if (selected_prev != selected){
    res();
  }

  selected_prev = selected;

  plot.display();
  mat.display();

  zin.display();
  zout.display();

  // display matrix information

  if(zin.isClicked() && zoom_level < 3){
    ymin += 0.5;
    ymax -= 0.5;
    xmin += 0.5;
    xmax -= 0.5;
    updateGraph();
    zoom_level +=1;
  }

  if(zout.isClicked()){ 
    ymin -= 0.5;
    ymax += 0.5;
    xmin -= 0.5;
    xmax += 0.5;
    updateGraph();
    zoom_level -=1;
  }

  textSize(20);
  textAlign(LEFT,CENTER);

}

// function for interacting with Javascript 
void setMatrix(a, b, c, d) {
  mat.setMatrix(float(a),float(b),float(c),float(d));
  setEigenvaluesForDisplay(mat.evalue1, mat.evalue2);
  setEigenvectorsForDisplay(mat.evec1a, mat.evec1b, mat.evec2a, mat.evec2b);
  setDeterminantForDisplay(mat.determinant);
}

// manipulate graph
void updateGraph(){
    plot.setXBounds(xmin,xmax);
    plot.setYBounds(ymin,ymax);
    mat.graphUpdate();
}

// reset the matrix and display
void res(){
    xmin = xmin_orig;
    xmax = xmax_orig;
    ymin = ymin_orig;
    ymax = ymax_orig;
    updateGraph();
    zoom_level = 0;
    
    mat.setMatrix(1,0,0,1);
    A.setNumber("1");
    B.setNumber("0");
    C.setNumber("0");
    D.setNumber("1");
    setEigenvaluesForDisplay(mat.evalue1, mat.evalue2);
    setEigenvectorsForDisplay(mat.evec1a, mat.evec1b, mat.evec2a, mat.evec2b);
    setDeterminantForDisplay(mat.determinant);
}

// limit output string length
String substr(String string, int leng){
  if(string.length() > leng){
    string = string.substring(0,leng);
  }
  return string;
}


/**
 * Matrix
 *
 * This class creates a 2x2 matrix object, initialized with values
 * a  b
 * c  d
 * with the intent of displaying it and its transformations graphically.
 *
 * @author Sarah Greer
 * @see Plot
 *
 */

public class Matrix{
  Plot plot;        // plot object for display
  float a,b,c,d;    // matrix values
  float det;        // determinant (for calculation)
  float e1, e2;     // eigenvalues (for calculation)
  float[] one, two; // eigenvectors (for calculation)
  String determinant;   // string determinant (for display)
  String evalue1, evalue2;                  // string eigenvalues (for display)
  String evec1a, evec1b, evec2a, evec2b;    // string eigenvectors (for display)

  // variables used for plotting and calculations
  private Boolean isImaginary;
  private int nx;
  private float miny, minx, maxx, dy, dx;
  private float unitx, unity, sqt22x, sqt22y;
  private float cx, cy;
  private float[] map1, map2, map3, map4, map5, map6, map7, map8;

  public Matrix(Plot plot, float a, float b, float c, float d){
    this.plot = plot;
    graphUpdate();
    isImaginary = false;
    this.a = a;
    this.b = b;
    this.c = c;
    this.d = d;

    one = new float[2];
    two = new float[2];

    generateEval();
    l2Norm();

    plotGenerate();
  }

  public void setMatrix(float a, float b, float c, float d){
    this.a = a;
    this.b = b;
    this.c = c;
    this.d = d;
    generateEval();
    l2Norm();
    plotGenerate();
  }

  // normalize to the unit vector (for plotting)
  public void normalizeToOne(){
    float min1 = one[0];
    float min2 = two[0];
    if(one[1] < one[0]){ min1 = one[1]; } 
    if(two[1] < two[0]){ min2 = two[1]; } 

    one[0] /= min1;
    one[1] /= min1;

    two[0] /= min2;
    two[1] /= min2;
  }

  // L2 norm
  void l2Norm(){
    float onenorm = sqrt(one[0]*one[0] + one[1]*one[1]);
    float twonorm = sqrt(two[0]*two[0] + two[1]*two[1]);
    one[0] /= onenorm;
    one[1] /= onenorm;

    two[0] /= twonorm;
    two[1] /= twonorm;
  }

  // generate the eigenvalues and eigenvecotrs
  void generateEval(){
    float T = a + d;            // trace
    det = a*d - b*c;            // determinant
    determinant = substr(str(det),8);
    float inRoot = T*T/4 - det; // value inside root
    float L1, L2;               // temporary storage values 

    // check for imaginary eigenvalues;
    if (inRoot >= 0){ 
      isImaginary = false;
      L1 = T/2 + sqrt(inRoot);
      L2 = T/2 - sqrt(inRoot);
      e1 = L1;
      e2 = L2;
      evalue1 = substr(str(e1),8);
      evalue2 = substr(str(e2),8);
    } else {
      isImaginary = true;
      L1 = 0;
      L2 = 0;
      e1 = L1;
      e2 = L2;
      evalue1 = str(T/2) + " + " + substr(str(sqrt(-1*inRoot)),5) + "i";
      evalue2 = str(T/2) + " - " + substr(str(sqrt(-1*inRoot)),5) + "i";
    }

    // calculate eigenvalues/eigenvectors.
    // doing this with a 2x2 matrix is easy...
    // case 1
    if (c != 0){
      one[0] = L1 - d;
      one[1] = c;
      two[0] = L2 - d;
      two[1] = c;
      if( ! isImaginary ){
        evec1a = substr(str(one[0]),8);
        evec1b = substr(str(one[1]),8);
        evec2a = substr(str(two[0]),8);
        evec2b = substr(str(two[1]),8);
      } else {
        float ft = T/2-d;
        if (ft != 0){
            evec1a = substr(str(ft),6) + " + " 
                     + substr(str(sqrt(-inRoot)),5) + "i";
            evec2a = substr(str(ft),6) + " - " 
                     + substr(str(sqrt(-inRoot)),5) + "i";
        } else{
            evec1a = substr(str(sqrt(-inRoot)),8) + "i";
            evec2a = substr(str(sqrt(-inRoot)),8) + "i";
        }
        evec1b = substr(str(c),8);
        evec2b = substr(str(c),8);
      }

    // case 2
    } else if (b != 0){
      one[0] = b;
      one[1] = L1 - a;
      two[0] = b;
      two[1] = L2 - a;
      if( ! isImaginary ){
        evec1a = substr(str(one[0]),8);
        evec1b = substr(str(one[1]),8);
        evec2a = substr(str(two[0]),8);
        evec2b = substr(str(two[1]),8);
      } else {
        float ft = T/2-a;
        if (ft != 0){
            evec1a = substr(str(ft),6) + " + " 
                     + substr(str(sqrt(-inRoot)),5) + "i";
            evec2a = substr(str(ft),6) + " - " 
                     + substr(str(sqrt(-inRoot)),5) + "i";
        } else{
            evec1a = substr(str(sqrt(-inRoot)),8) + "i";
            evec2a = substr(str(sqrt(-inRoot)),8) + "i";
        }
        evec1a = substr(str(b),8);
        evec2a = substr(str(b),8);
      }
    // case 3
    } else {
      one[0] = 1;
      one[1] = 0;
      two[0] = 0;
      two[1] = 1;
        evec1a = substr(str(one[0]),8);
        evec1b = substr(str(one[1]),8);
        evec2a = substr(str(two[0]),8);
        evec2b = substr(str(two[1]),8);
    }
  }
  
  // update the graph
  void graphUpdate(){
    nx = plot.nx;
    miny = plot.miny;
    minx = plot.minx;
    maxx = plot.maxx;
    dy = plot.dy;
    dx = plot.dx;
    unitx = 1/dx;
    unity = 1/dy;
    sqt22x = 1/dx * sqrt(2)/2;
    sqt22y = 1/dy * sqrt(2)/2;
    cx = plot.cx;
    cy = plot.cy;
    plotGenerate();

  }

  // generate the plot
  void plotGenerate(){
    // update
    map1 = transA(-unitx,0);
    map2 = transA(-sqt22x,sqt22y);
    map3 = transA(0,unity);
    map4 = transA(sqt22x,sqt22y);
    map5 = transA(unitx,0);
    map6 = transA(sqt22x,-sqt22y);
    map7 = transA(0,-unity);
    map8 = transA(-sqt22x,-sqt22y);
  }

  // apply the transformation on the given vector
  float[] transA(float x, float y){
    float x1 = a*x+b*y;
    float y1 = c*x+d*y;
    return new float[]{ x1, y1 };
  }

  // display the result
  void display(){
    // create ellipse
    // I do this by plotting 8 points and interpolating between them.
    // It's not perfect, but it works well enough
    // Alternatively, I could easily plot a square...
    noStroke();
    fill(255,255,0,210);
    beginShape();
        curveVertex(cx + map1[0], cy - map1[1]);
        curveVertex(cx + map1[0], cy - map1[1]);
        curveVertex(cx + map2[0], cy - map2[1]);
        curveVertex(cx + map3[0], cy - map3[1]);
        curveVertex(cx + map4[0], cy - map4[1]);
        curveVertex(cx + map5[0], cy - map5[1]);
        curveVertex(cx + map6[0], cy - map6[1]);
        curveVertex(cx + map7[0], cy - map7[1]);
        curveVertex(cx + map8[0], cy - map8[1]);
        curveVertex(cx + map1[0], cy - map1[1]);
        curveVertex(cx + map2[0], cy - map2[1]);
        curveVertex(cx + map3[0], cy - map3[1]);
        curveVertex(cx + map3[0], cy - map3[1]);
    endShape();

    // create eigenvectors
    strokeWeight(7.5);
    stroke(#00ff00);
    line(cx,cy, (cx + e1*one[0]*unitx), (cy - e1*one[1]*unity));
    stroke(#0000ff);
    line(cx,cy, (cx + e2*two[0]*unitx), (cy - e2*two[1]*unity));
    strokeWeight(1);
  }

}


/** 
 * Button 
 *
 * This class creates a button object with center position (xpos, ypos), width 
 * xsize, height ysize, all in pixels, and color c.
 *
 * @author Sarah Greer
 * @see TextBox
 * @see RadioButton
 * @see SmallButton
 *
 */

public class Button {
  color c, c2, c3;      // colors for fill, stroke, and text
  float xpos, ypos;     // center x and y of button
  float xsize, ysize;   // width and height of button
  float x2, y2;         // half width and height of button (for borders)
  int sw;               // stroke weight
  String txt = "";      // button text
  float txtsize;        // text size
  color hoverc = color(0,0,0,50);       // shaded darker when hovered
  color clickc = color(0,0,0,125);      // shaded darker when clicked
  
  public Button(float xpos, float ypos, float xsize, float ysize, color c) {
    this.xpos = xpos;
    this.ypos = ypos;
    this.xsize = xsize;
    this.ysize = ysize;
    this.c = c;
    x2 = xsize/2;
    y2 = ysize/2;

    // default stroke color and weight 
    c2 = #000000;
    sw = 3;
  }

  // set fill color
  protected void setF(color c){
    this.c = c;
  }

  // set stroke color
  protected void setS(color c){
    this.c2 = c;
  }

  // set stroke weight
  protected void setStrokeWeight(int sw){
    this.sw = sw;
  }

  // set button text
  protected void setText(int txtsize, String txt, color c){
    this.txtsize = txtsize;
    textSize(txtsize);
    this.txt = txt;
    this.c3 = c;
  }
  
  public void display() {
    stroke(c2);
    strokeWeight(sw);

    fill(c);
    rectMode(CENTER);
    rect(xpos,ypos,xsize,ysize);

    // fill with clicked color
    if ( isClicked() ){ 
      fill(clickc);
      rect(xpos,ypos,xsize,ysize);
    
    // fill with hover color
    } else if ( isHovering() ){
      fill(hoverc);
      rect(xpos,ypos,xsize,ysize);
    }

    // add text if there is any
    if (txt != ""){
      textSize(txtsize);
      fill (c3);
      textAlign(CENTER, CENTER);
      text(txt, xpos, ypos);
    }
  }
  
  // checks if button is being hovered over
  protected boolean isHovering(){
    if (mouseX > xpos-x2 && mouseX < xpos+x2 &&
        mouseY > ypos-y2 && mouseY < ypos+y2) return true; 
    else return false; 
  }

  // checks if button is clicked
  protected boolean isClicked() {
    if (isHovering() && mousePressed) return true; 
    else return false; 
  }
}


/**
 * Plot
 *
 * This class creates a plot object which can be used to display functions
 * (Line, VL, Vert classes) or matrix representations (matrix class).
 *
 * Currently has pan and zoom functionality (by changing x and y bounds) 
 *
 * Panning functionality, useful for function display, is only partially 
 * supported.
 *
 * @author Sarah Greer
 * @see Matrix
 * @see Line
 * @see VL
 * @see Vert
 *
 */

public class Plot{
  int nx, ny;   // number of x and y pixels
  int cx, cy;   // center of x and y
  color c;      // selected colors
  float minx, miny, maxx, maxy;     // plot x and y bounds
  float[] x, y;         // coordinates mapped to pixels
  float[] xmap, ymap;   // coordinated mapped to real values
  float dx, dy;         // spacing for plotting units

  public Plot(int nx, int ny, float minx, float miny, float maxx, float maxy, 
      color c){
    this.c = c;
    this.nx = nx;
    this.ny = ny;
    cx = nx/2;
    cy = ny/2;

    this.minx = minx;
    this.miny = miny;
    this.maxx = maxx;
    this.maxy = maxy;

    x = new float[nx];
    y = new float[ny];

    xmap = new float[nx];
    ymap = new float[ny];

    updatePlot();
  }

  // updates the display
  private void updatePlot(){
    dx = (maxx - minx) / nx;
    dy = (maxy - miny) / ny;

    for (int i = 0; i < nx; i++){
      x[i] = minx + dx*i;
      xmap[i] = i;
    }
    for (int i = 0; i < ny; i++){
      y[i] = miny + dy*i;
      ymap[i] = i;
    }
  }

  // sets x bounds
  public void setXBounds(float minx, float maxx){
    this.minx = minx;
    this.maxx = maxx;
    updatePlot();
  }

  // sets y bounds
  public void setYBounds(float miny, float maxy){
    this.miny = miny;
    this.maxy = maxy;
    updatePlot();
  }

  // changes background color for plot
  public void changeColor(color c){
    c = c;
  }

  // display the plot
  public void display(){
    rectMode(CENTER);
    fill(c);
    stroke(c);
    rect(cx,cy,nx,ny);

    // write the bounds
    stroke(#696969);
    fill(#696969);
    text(str(ymax), cx+10, ymap[0]+15);
    text(str(ymin), cx+10, ymap[ny-1]-15);
    text(str(xmin), xmap[0]+10, cy+20);
    text(str(xmax), xmap[nx-1]-35, cy+20);

    // set graph lines (for panning -- partially implemented)
    int y0ind = cy, x0ind = cx;
    float y0err = 0.1, x0err = 0.1;
    for (int i = 1; i < ny; i++){
      if (y[i] >= 0){
        if (y[i] < y0err){ 
          y0err = y[i];
          y0ind = i; 
        }
      } 
      if (i%100 == 0){
        strokeWeight(1);
        line(0-2,ymap[i],nx+2,ymap[i]);
      } else if (i%50 == 0){
        strokeWeight(1);
        line(nx-2,ymap[i],nx+2,ymap[i]);
        line(0-2,ymap[i],0+2,ymap[i]);
      }
    }

    // set graph lines
    strokeWeight(5);
    line(0-2,ymap[y0ind],nx+2,ymap[y0ind]);
    for (int i = 1; i < nx; i++){
      if (x[i] >= 0){
        if (x[i] < x0err){  
          x0err = x[i];
          x0ind = i; 
        }
      } 
      if (i%100 == 0){
        strokeWeight(1);
        line(xmap[i],0-2,xmap[i],ny+2);
      } else if (i%50 == 0){
        strokeWeight(1);
        line(xmap[i],ny-2,xmap[i],ny+2);
        line(xmap[i],0-2,xmap[i],0+2);
      }
    }

    strokeWeight(5);
    line(xmap[x0ind],0-2,xmap[x0ind],ny+2);
    strokeWeight(1);
  }
}
